#include "zpfft.h"

#define reg register double
#define V *(volatile const double *)&

void zpfft_2(const zpfft_p *c,double *u)
{
  reg t0, t1;

  t0 = u[0];
  t0 += u[1];
  t1 = u[0];
  t1 -= u[1];

  u[0] = t0;
  u[1] = t1;
}

static inline void generic(register const zpfft_p *c,register double *u,register const int n)
{
  register int loop;
  register int itimes256n;
  reg t0, t1, t2, t3, t4, t5, t6, t7;

  loop = n;
  itimes256n = 0;

  do {
    t4 = u[0];
    t4 -= u[2 * n];
    t6 = u[n];
    t6 -= u[3 * n];
    t4 *= *(const double *) (itimes256n * (sizeof(double) / 2) + (const char *) c->roots);
    t0 = u[0];
    t0 += u[2 * n];
    t6 *= *(const double *) (128 * sizeof(double) + itimes256n * (sizeof(double) / 2) + (const char *) c->roots);
    t5 = V c->pinv;
    t5 *= t4;
    t7 = V c->pinv;
    t7 *= t6;
    t2 = u[n];
    t2 += u[3 * n];
    t5 += c->alpha0;
    t7 += c->alpha0;
    t1 = t0;
    t5 -= c->alpha0;
    t7 -= c->alpha0;
    t0 += t2;
    t5 *= c->p;
    t2 = t1 - t2;
    t7 *= c->p;
    t4 -= t5;
    t2 *= c->roots[itimes256n];
    t6 -= t7;
    t5 = t4;
    t1 = V c->pinv;
    t4 += t6;
    t1 *= t0;
    t6 = t5 - t6;
    t3 = V c->pinv;
    t3 *= t2;
    t1 += c->alpha0;
    t6 *= c->roots[itimes256n];
    t7 = V c->pinv;
    t7 *= t6;
    t3 += c->alpha0;
    t1 -= c->alpha0;
    t7 += c->alpha0;
    t3 -= c->alpha0;
    t1 *= c->p;
    t7 -= c->alpha0;
    t3 *= c->p;
    itimes256n += 256 / n;
    ++u;
    t7 *= c->p;
    t0 -= t1;
    t2 -= t3;
    t6 -= t7;
    u[-1 + 2 * n] = t4;
    u[-1] = t0;
    u[-1 + n] = t2;
    u[-1 + 3 * n] = t6;
  }
  while (--loop);
}

inline void zpfft_4(register const zpfft_p *c,register double *u)
{
  reg t0, t1, t2, t3, t4, t5, t6, t7;

  t0 = u[0];
  t0 += u[2];
  t2 = u[1];
  t2 += u[3];
  t6 = u[1];
  t6 -= u[3];
  t4 = u[0];
  t4 -= u[2];
  t6 *= c->roots[128];
  t1 = t0;
  t0 += t2;
  t2 = t1 - t2;
  t5 = t4;
  t4 += t6;
  t6 = t5 - t6;
  t1 = V c->pinv;
  t1 *= t0;
  t3 = V c->pinv;
  t3 *= t2;
  t1 += c->alpha0;
  t5 = V c->pinv;
  t5 *= t4;
  t3 += c->alpha0;
  t1 -= c->alpha0;
  t7 = V c->pinv;
  t7 *= t6;
  t3 -= c->alpha0;
  t5 += c->alpha0;
  t1 *= c->p;
  t7 += c->alpha0;
  t5 -= c->alpha0;
  t3 *= c->p;
  t7 -= c->alpha0;
  t5 *= c->p;
  t0 -= t1;
  t7 *= c->p;
  t4 -= t5;
  t2 -= t3;
  t6 -= t7;
  u[0] = t0;
  u[1] = t2;
  u[2] = t4;
  u[3] = t6;
}

void zpfft_8(const zpfft_p *c,double *u)
{
  reg t0, t1, t2, t3;

  generic(c,u,2);

  t0 = u[0];
  t0 += u[1];
  t1 = u[0];
  t1 -= u[1];
  t2 = u[2];
  t2 += u[3];
  t3 = u[2];
  t3 -= u[3];
  u[0] = t0;
  u[2] = t2;
  u[1] = t1;
  u[3] = t3;

  t0 = u[4];
  t0 += u[5];
  t1 = u[4];
  t1 -= u[5];
  t2 = u[6];
  t2 += u[7];
  t3 = u[6];
  t3 -= u[7];
  u[4] = t0;
  u[6] = t2;
  u[5] = t1;
  u[7] = t3;
}

void zpfft_16(const zpfft_p *c,double *u)
{
  generic(c,u,4);
  zpfft_4(c,u);
  zpfft_4(c,u + 4);
  zpfft_4(c,u + 8);
  zpfft_4(c,u + 12);
}

void zpfft_32(const zpfft_p *c,double *u)
{
  generic(c,u,8);
  zpfft_8(c,u);
  zpfft_8(c,u + 8);
  zpfft_8(c,u + 16);
  zpfft_8(c,u + 24);
}

void zpfft_64(const zpfft_p *c,double *u)
{
  generic(c,u,16);
  zpfft_16(c,u);
  zpfft_16(c,u + 16);
  zpfft_16(c,u + 32);
  zpfft_16(c,u + 48);
}

void zpfft_128(const zpfft_p *c,double *u)
{
  generic(c,u,32);
  zpfft_32(c,u);
  zpfft_32(c,u + 32);
  zpfft_32(c,u + 64);
  zpfft_32(c,u + 96);
}

void zpfft_256(const zpfft_p *c,double *u)
{
  generic(c,u,64);
  zpfft_64(c,u);
  zpfft_64(c,u + 64);
  zpfft_64(c,u + 128);
  zpfft_64(c,u + 192);
}

void zpfft_512(const zpfft_p *c,double *u)
{
  generic(c,u,128);
  zpfft_128(c,u);
  zpfft_128(c,u + 128);
  zpfft_128(c,u + 256);
  zpfft_128(c,u + 384);
}
