#include "zpfft.h"

#define reg register double
#define V *(volatile const double *)&

static inline void generic_un(register const zpfft_p *c,register double *u,register const int n)
{
  register int loop;
  register int itimes256n;
  reg t0, t1, t2, t3, t4, t5, t6, t7;

  loop = n;
  itimes256n = 256;

  do {
    t6 = u[3 * n];
    t6 *= c->roots[itimes256n];
    t2 = u[n];
    t7 = V c->pinv;
    t7 *= t6;
    t0 = u[0];
    t2 *= c->roots[itimes256n];
    t7 += c->alpha0;
    t4 = u[2 * n];
    t0 -= t2;
    t7 -= c->alpha0;
    t2 += u[0];
    t1 = t0;
    t7 *= c->p;
    t3 = t2;
    t5 = V c->pinv;
    t6 -= t7;
    itimes256n -= 256 / n;
    ++u;
    t7 = V c->pinv;
    t4 -= t6;
    t6 += u[-1 + 2 * n];

    t4 *= *(const double *) ((256 + 256 / n) * (sizeof(double) / 2) + itimes256n * (sizeof(double) / 2) + (const char *) c->roots);

    t6 *= *(const double *) ((256 / n) * (sizeof(double) / 2) + itimes256n * (sizeof(double) / 2) + (const char *) c->roots);
    t0 -= t4;
    t4 += t1;
    t2 -= t6;
    t6 += t3;
    t5 *= t4;
    t1 = V c->pinv;
    t1 *= t0;
    t5 += c->alpha0;
    t7 *= t6;
    t1 += c->alpha0;
    t3 = V c->pinv;
    t3 *= t2;
    t5 -= c->alpha0;
    t1 -= c->alpha0;
    t3 += c->alpha0;
    t5 *= c->p;
    t7 += c->alpha0;
    t3 -= c->alpha0;
    t1 *= c->p;
    t7 -= c->alpha0;
    t3 *= c->p;
    t0 -= t1;
    t7 *= c->p;
    t4 -= t5;
    t2 -= t3;
    t6 -= t7;
    u[-1] = t0;
    u[-1 + 2 * n] = t4;
    u[-1 + 3 * n] = t6;
    u[-1 + n] = t2;
  }
  while (--loop);
}

inline void zpfft_un4(register const zpfft_p *c,register double *u)
{
  reg t0, t1, t2, t3, t4, t5, t6, t7;

  t6 = u[2];
  t6 -= u[3];
  t2 = u[0];
  t2 -= u[1];
  t6 *= c->roots[128];
  t4 = u[2];
  t4 += u[3];
  t0 = u[0];
  t0 += u[1];
  t3 = t2;
  t2 -= t6;
  t6 += t3;
  t1 = t0;
  t0 += t4;
  t4 = t1 - t4;
  t3 = V c->pinv;
  t3 *= t2;
  t7 = V c->pinv;
  t7 *= t6;
  t5 = V c->pinv;
  t5 *= t4;
  t1 = V c->pinv;
  t1 *= t0;
  t3 += c->alpha0;
  t7 += c->alpha0;
  t5 += c->alpha0;
  t3 -= c->alpha0;
  t1 += c->alpha0;
  t7 -= c->alpha0;
  t3 *= c->p;
  t5 -= c->alpha0;
  t7 *= c->p;
  t1 -= c->alpha0;
  t5 *= c->p;
  t2 -= t3;
  t1 *= c->p;
  t6 -= t7;
  t4 -= t5;
  t0 -= t1;
  u[1] = t2;
  u[3] = t6;
  u[2] = t4;
  u[0] = t0;
}

void zpfft_un8(const zpfft_p *c,double *u)
{
  reg t0, t1, t2, t3, t4, t5, t6, t7;

  t0 = u[0];
  t0 += u[1];
  t1 = u[0];
  t1 -= u[1];
  t2 = u[2];
  t2 += u[3];
  t3 = u[2];
  t3 -= u[3];
  t4 = u[4];
  t4 += u[5];
  t5 = u[4];
  t5 -= u[5];
  t6 = u[6];
  t6 += u[7];
  t7 = u[6];
  t7 -= u[7];

  u[0] = t0;
  u[1] = t1;
  u[2] = t2;
  u[3] = t3;
  u[4] = t4;
  u[5] = t5;
  u[6] = t6;
  u[7] = t7;

  generic_un(c,u,2);
}

void zpfft_un16(const zpfft_p *c,double *u)
{
  zpfft_un4(c,u);
  zpfft_un4(c,u + 4);
  zpfft_un4(c,u + 8);
  zpfft_un4(c,u + 12);
  generic_un(c,u,4);
}

void zpfft_un32(const zpfft_p *c,double *u)
{
  zpfft_un8(c,u);
  zpfft_un8(c,u + 8);
  zpfft_un8(c,u + 16);
  zpfft_un8(c,u + 24);
  generic_un(c,u,8);
}

void zpfft_un64(const zpfft_p *c,double *u)
{
  zpfft_un16(c,u);
  zpfft_un16(c,u + 16);
  zpfft_un16(c,u + 32);
  zpfft_un16(c,u + 48);
  generic_un(c,u,16);
}

void zpfft_un128(const zpfft_p *c,double *u)
{
  zpfft_un32(c,u);
  zpfft_un32(c,u + 32);
  zpfft_un32(c,u + 64);
  zpfft_un32(c,u + 96);
  generic_un(c,u,32);
}

void zpfft_un256(const zpfft_p *c,double *u)
{
  zpfft_un64(c,u);
  zpfft_un64(c,u + 64);
  zpfft_un64(c,u + 128);
  zpfft_un64(c,u + 192);
  generic_un(c,u,64);
}

void zpfft_un512(const zpfft_p *c,double *u)
{
  zpfft_un128(c,u);
  zpfft_un128(c,u + 128);
  zpfft_un128(c,u + 256);
  zpfft_un128(c,u + 384);
  generic_un(c,u,128);
}
